<?php
/**
 * The model file of feedback module of ZenTaoCMS.
 *
 * @copyright   Copyright 2009-2015 青岛易软天创网络科技有限公司(QingDao Nature Easy Soft Network Technology Co,LTD, www.cnezsoft.com)
 * @license     ZPL (http://zpl.pub/page/zplv12.html)
 * @author      Yidong Wang <yidong@cnezsoft.com>
 * @package     feedback
 * @version     $Id$
 * @link        http://www.zentao.net
 */
class feedbackModel extends model
{
    public function create()
    {
        $needReview = !$this->forceNotReview();
        if($needReview)
        {
            $assignedTo = $this->config->feedback->reviewer;
            if(empty($assignedTo)) $assignedTo = $this->loadModel('dept')->getManager($this->app->user->dept);
        }
        else
        {
            $assignedTo = $this->dao->findByID($this->post->product)->from(TABLE_PRODUCT)->fetch('feedback');
        }

        $now      = helper::now();
        $status   = $this->getStatus('create', $needReview);
        $feedback = fixer::input('post')
            ->add('openedBy', $this->app->user->account)
            ->add('openedDate', $now)
            ->add('editedBy', $this->app->user->account)
            ->add('editedDate', $now)
            ->add('status', $status)
            ->setDefault('module', 0)
            ->setIF(isset($assignedTo), 'assignedTo', $assignedTo)
            ->setIF(isset($assignedTo), 'assignedDate', $now)
            ->setDefault('product', 0)
            ->join('mailto', ',')
            ->stripTags($this->config->feedback->editor->create['id'])
            ->remove('files,labels,uid')
            ->get();

        $feedback = $this->loadModel('file')->processImgURL($feedback, $this->config->feedback->editor->create['id'], $this->post->uid);
        $this->dao->insert(TABLE_FEEDBACK)->data($feedback)->autoCheck()->batchCheck($this->config->feedback->create->requiredFields, 'notempty')->exec();
        if(!dao::isError())
        {
            $feedbackID = $this->dao->lastInsertID();
            $this->file->updateObjectID($this->post->uid, $feedbackID, 'feedback');
            $this->file->saveUpload('feedback', $feedbackID);
            return $feedbackID;
        }
        return false;
    }

    public function update($id)
    {
        $oldFeedback = $this->dao->select('*')->from(TABLE_FEEDBACK)->where('id')->eq((int)$id)->fetch();
        if(strpos('noreview|wait|clarify', $oldFeedback->status) === false) return false;

        $needReview = !$this->forceNotReview();
        $status     = $this->getStatus('update', $needReview, $oldFeedback->status);
        if($status == 'noreview')
        {
            $assignedTo = $this->config->feedback->reviewer;
            if(empty($assignedTo)) $assignedTo = $this->loadModel('dept')->getManager($this->app->user->dept);
        }
        elseif($this->post->product != '' and $status == 'wait')
        {
            $assignedTo = $this->dao->findByID($this->post->product)->from(TABLE_PRODUCT)->fetch('feedback');
        }

        $now        = helper::now();
        $feedback   = fixer::input('post')
            ->add('editedBy', $this->app->user->account)
            ->add('editedDate', $now)
            ->add('status', $status)
            ->setDefault('module', 0)
            ->setIF((isset($assignedTo) && $assignedTo != $oldFeedback->assignedTo), 'assignedTo', $assignedTo)
            ->setIF((isset($assignedTo) && $assignedTo != $oldFeedback->assignedTo), 'assignedDate', $now)
            ->setIF($this->post->public != 1, 'public', 0)
            ->setIF($this->post->notify != 1, 'notify', 0)
            ->stripTags($this->config->feedback->editor->edit['id'])
            ->join('mailto', ',')
            ->remove('files,labels,uid')
            ->get();

        $feedback = $this->loadModel('file')->processImgURL($feedback, $this->config->feedback->editor->edit['id'], $this->post->uid);
        $this->dao->update(TABLE_FEEDBACK)->data($feedback)->autoCheck()->batchCheck($this->config->feedback->edit->requiredFields, 'notempty')->where('id')->eq($id)->exec();
        if(!dao::isError())
        {
            $this->file->updateObjectID($this->post->uid, $id, 'feedback');
            return common::createChanges($oldFeedback, $feedback);
        }
        return false;
    }

    public function batchUpdate()
    {
        $data = fixer::input('post')->get();
        $oldFeedbacks = $this->getByList(array_keys($data->titles));

        $now     = helper::now();
        $changes = array();
        foreach($data->titles as $feedbackID => $title)
        {
            $oldFeedback = $oldFeedbacks[$feedbackID];
            $feedback = new stdclass();
            $feedback->title      = $title;
            $feedback->editedBy   = $this->app->user->account;
            $feedback->editedDate = $now;
            $feedback->product    = $data->products[$feedbackID];
            $feedback->module     = $data->modules[$feedbackID];
            $feedback->assignedTo = $data->assignedTos[$feedbackID];
            if($feedback->assignedTo != $oldFeedback->assignedTo) $feedback->assignedDate = $now;
            $this->dao->update(TABLE_FEEDBACK)->data($feedback)->where('id')->eq($feedbackID)->exec();
            $changes[$feedbackID] = common::createChanges($oldFeedback, $feedback);
        }

        return $changes;
    }

    public function assign($feedbackID)
    {
        $now = helper::now();
        $oldFeedback = $this->getById($feedbackID);
        $feedback = fixer::input('post')
            ->setDefault('editedBy', $this->app->user->account)
            ->setDefault('editedDate', $now)
            ->setDefault('assignedDate', $now)
            ->join('mailto', ',')
            ->remove('comment,contactListMenu')
            ->get();

        $this->dao->update(TABLE_FEEDBACK)
            ->data($feedback)
            ->autoCheck()
            ->where('id')->eq($feedbackID)->exec();

        if(!dao::isError()) return common::createChanges($oldFeedback, $feedback);
    }

    public function batchAssign()
    {
        $oldFeedbacks = $this->getByList($this->post->feedbackIDList);

        $now     = helper::now();
        $changes = array();
        foreach($this->post->feedbackIDList as $feedbackID)
        {
            $oldFeedback = $oldFeedbacks[$feedbackID];
            if($oldFeedback->assignedTo == $this->post->assignedTo) continue;

            $feedback = new stdclass();
            $feedback->assignedTo   = $this->post->assignedTo;
            $feedback->assignedDate = $now;
            $feedback->editedBy     = $this->app->user->account;
            $feedback->editedDate   = $now;

            $this->dao->update(TABLE_FEEDBACK)->data($feedback)->where('id')->eq($feedbackID)->exec();
            $changes[$feedbackID] = common::createChanges($oldFeedback, $feedback);
        }
        return $changes;
    }

    /**
     * Batch change the module of feedback.
     *
     * @param  array  $feedbackIDList
     * @param  int    $moduleID
     * @access public
     * @return array
     */
    public function batchChangeModule($feedbackIDList, $moduleID)
    {
        $now          = helper::now();
        $allChanges   = array();
        $oldFeedbacks = $this->getByList($feedbackIDList);
        foreach($feedbackIDList as $feedbackID)
        {
            $oldFeedback = $oldFeedbacks[$feedbackID];
            if($moduleID == $oldFeedback->module) continue;

            $feedback = new stdclass();
            $feedback->module = $moduleID;

            $this->dao->update(TABLE_FEEDBACK)->data($feedback)->autoCheck()->where('id')->eq((int)$feedbackID)->exec();
            if(!dao::isError()) $allChanges[$feedbackID] = common::createChanges($oldFeedback, $feedback);
        }
        return $allChanges;
    }

    public function review($feedbackID)
    {
        if($this->post->result == false) die(js::alert($this->lang->feedback->mustChooseResult));
        $now          = helper::now();
        $oldFeedback  = $this->dao->findById($feedbackID)->from(TABLE_FEEDBACK)->fetch();
        $status       = $this->getStatus('review', true);
        if($this->post->result == 'pass')
        {
            $assignedTo   = $this->dao->findByID($oldFeedback->product)->from(TABLE_PRODUCT)->fetch('feedback');
            $assignedDate = $now;
        }
        if($this->post->result == 'clarify')
        {
            $assignedTo   = $oldFeedback->openedBy;
            $assignedDate = $now;
        }

        $feedback = fixer::input('post')
            ->remove('result,comment')
            ->setDefault('reviewedDate', $now)
            ->setDefault('editedBy', $this->app->user->account)
            ->setDefault('editedDate', $now)
            ->setIF($status, 'status', $status)
            ->setIF($status, 'assignedTo', $assignedTo)
            ->setIF($status, 'assignedDate', $assignedDate)
            ->join('reviewedBy', ',')
            ->get();

        $this->dao->update(TABLE_FEEDBACK)->data($feedback)->autoCheck()->where('id')->eq($feedbackID)->exec();
        if(dao::isError()) return false;

        return common::createChanges($oldFeedback, $feedback);
    }

    public function batchReview($feedbackIdList, $result)
    {
        if(empty($feedbackIdList)) return true;

        $now     = helper::now();
        $actions = array();
        $this->loadModel('action');

        $oldFeedbacks = $this->dao->select('*')->from(TABLE_FEEDBACK)->where('id')->in($feedbackIdList)->fetchAll('id');
        foreach($oldFeedbacks as $oldFeedback) $products[$oldFeedback->product] = $oldFeedback->product;
        $productFeedbacks = $this->dao->select('id,feedback')->from(TABLE_PRODUCT)->where('id')->in($products)->fetchPairs('id', 'feedback');
        foreach($feedbackIdList as $feedbackID)
        {
            $oldFeedback = $oldFeedbacks[$feedbackID];
            if($oldFeedback->status != 'noreview') continue;

            $feedback = new stdClass();
            $feedback->reviewedBy   = $this->app->user->account;
            $feedback->reviewedDate = $now;
            $feedback->editedBy     = $this->app->user->account;
            $feedback->editedDate   = $now;
            if($result == 'pass')
            {
                $feedback->status = 'wait';
                if(isset($productFeedbacks[$oldFeedback->product]))
                {
                    $feedback->assignedTo   = $productFeedbacks[$oldFeedback->product];
                    $feedback->assignedDate = $now;
                }
            }
            if($result == 'clarify')
            {
                $feedback->status       = 'clarify';
                $feedback->assignedTo   = $oldFeedback->openedBy;
                $feedback->assignedDate = $now;
            }
            $this->dao->update(TABLE_FEEDBACK)->data($feedback)->autoCheck()->where('id')->eq($feedbackID)->exec();
            $actions[$feedbackID] = $this->action->create('feedback', $feedbackID, 'Reviewed', '', ucfirst($result));
        }

        return $actions;
    }

    public function like($feedbackID)
    {
        $feedback = $this->dao->select('id, likes')->from(TABLE_FEEDBACK)->where('id')->eq($feedbackID)->fetch();

        $likeBy = ',' . $this->app->user->account . ',';
        $likes  = ',' . trim($feedback->likes, ',') . ',';
        if(strpos($likes, $likeBy) !== false)
        {
            $likes = trim(str_replace($likeBy, '', $likes), ',');
        }
        else
        {
            $likes = ltrim($likes, ',') . $this->app->user->account;
        }
        $this->dao->update(TABLE_FEEDBACK)->set('likes')->eq($likes)->where('id')->eq($feedbackID)->exec();

        return $likes;
    }

    public function getList($browseType = 'unclosed', $orderBy = 'id_desc', $pager = 'null', $param = 0)
    {
        $grantProducts = $this->getGrantProducts();
        return $this->dao->select('*')->from(TABLE_FEEDBACK)->where('deleted')->eq('0')
            ->beginIF($browseType == 'unclosed')->andWhere('status')->ne('closed')->fi()
            ->beginIF($browseType == 'public')->andWhere('public')->eq('1')->fi()
            ->beginIF($browseType == 'tostory')->andWhere('closedReason')->eq('tostory')->fi()
            ->beginIF($browseType == 'tobug')->andWhere('closedReason')->eq('tobug')->fi()
            ->beginIF($browseType == 'totask')->andWhere('closedReason')->eq('totask')->fi()
            ->beginIF($browseType == 'totodo')->andWhere('closedReason')->eq('totodo')->fi()
            ->beginIF($browseType == 'assigntome')->andWhere('assignedTo')->eq($this->app->user->account)->fi()
            ->beginIF($browseType == 'byModule')->andWhere('module')->eq((int)$param)->fi()
            ->beginIF($browseType != 'public' and $this->app->getMethodName() == 'browse')->andWhere('openedBy')->eq($this->app->user->account)->fi()
            ->beginIF($browseType != 'public')
            ->andWhere('product', true)->in(array_keys($grantProducts))
            ->orWhere('assignedTo')->eq($this->app->user->account)
            ->markRight(1)
            ->beginIF($browseType == 'review')
            ->andWhere('status')->in('noreview')
            ->fi()
            ->beginIF($browseType != 'review')
            ->andWhere('status', true)->notIN('noreview,clarify')
            ->orWhere('openedBy')->eq($this->app->user->account)
            ->markRight(1)
            ->fi()
            ->fi()
            ->orderBy($orderBy)
            ->page($pager)
            ->fetchAll();
    }

    public function getById($feedbackID)
    {
        $feedback = $this->dao->select('*')->from(TABLE_FEEDBACK)
            ->where('id')->eq($feedbackID)
            ->fetch();

        $feedback->likesCount = $feedback->likes ? count(explode(',', $feedback->likes)) : 0;
        $feedback->files      = $this->loadModel('file')->getByObject('feedback', $feedbackID); 

        $feedback = $this->file->replaceImgURL($feedback, 'desc'); 
        $feedback->desc = $this->file->setImgSize($feedback->desc); 

        return $feedback;
    }

    public function getByList($idList)
    {
        return $this->dao->select('*')->from(TABLE_FEEDBACK)->where('id')->in($idList)->fetchAll('id');
    }

    public function getBySearch($queryID = 0, $orderBy = 'id_desc', $pager = null)
    {
        $methodName    = $this->app->methodName;
        $grantProducts = $this->getGrantProducts();
        $this->loadModel('search');
        $query = $queryID ? $this->search->getQuery($queryID) : '';
        if($query)
        {
            $this->session->set('feedbackQuery', $query->sql);
            $this->session->set('feedbackForm', $query->form);
        }
        if($this->session->feedbackQuery == false) $this->session->set('feedbackQuery', ' 1 = 1');

        $feedbackQuery = $this->session->feedbackQuery;

        return $this->dao->select('*')->from(TABLE_FEEDBACK)
            ->where($feedbackQuery)
            ->andWhere('deleted')->eq('0')
            ->beginIF($methodName == 'admin')
            ->andWhere('product', true)->in(array_keys($grantProducts))
            ->orWhere('assignedTo')->eq($this->app->user->account)
            ->markRight(1)
            ->fi()
            ->beginIF($methodName == 'browse')
            ->andWhere('public', true)->eq('1')
            ->orWhere('openedBy')->eq($this->app->user->account)
            ->markRight(1)
            ->fi()
            ->andWhere('status', true)->notIN('noreview,clarify')
            ->orWhere('openedBy')->eq($this->app->user->account)
            ->markRight(1)
            ->orderBy($orderBy)
            ->page($pager)
            ->fetchAll();
    }

    public function getFeedbackPairs($type = 'feedback')
    {
        $userPairs = $this->dao->select('*')->from(TABLE_USER)->where('deleted')->eq(0)
            ->beginIF($type == 'feedback')->andWhere('feedback')->eq(1)->fi()
            ->beginIF($type != 'feedback')->andWhere('feedback')->eq(0)->fi()
            ->fetchPairs('account', 'realname');

        foreach($userPairs as $account => $realname)
        {
            if(empty($realname))$userPairs[$account] = $account;
        }
        return $userPairs;
    }

    public function getUserFeedbackPairs($account = '', $limit = 10)
    {
        $feedbackPairs = $this->dao->select('id, title')->from(TABLE_FEEDBACK)
            ->where('deleted')->eq(0)
            ->andWhere('assignedTo')->eq($account)
            ->andWhere('status')->ne('noreview')
            ->andWhere('status')->ne('clarify')
            ->orderBy('id_desc')
            ->limit($limit)
            ->fetchPairs();

        return $feedbackPairs;
    }

    public function getFeedbackView($products)
    {
        return $this->dao->select('*')->from(TABLE_FEEDBACKVIEW)
            ->beginIF($products)->where('product')->in($products)->fi()
            ->fetchGroup('product', 'account');
    }

    public function getGrantProducts()
    {
        $products = $this->dao->select('*')->from(TABLE_FEEDBACKVIEW)->where('account')->eq($this->app->user->account)->fetchPairs('product', 'product');
        $allottedProducts = array();
        if(empty($products)) $allottedProducts = $this->dao->select('DISTINCT product')->from(TABLE_FEEDBACKVIEW)->fetchPairs('product', 'product');
        $admin = $this->app->user->admin;
        return $this->dao->select('*')->from(TABLE_PRODUCT)
            ->where('deleted')->eq('0')
            ->beginIF(!empty($products) and !$admin)->andWhere('id')->in($products)->fi()
            ->beginIF(!empty($allottedProducts) and !$admin)->andWhere('id')->notin($allottedProducts)->fi()
            ->fetchPairs('id', 'name');
    }

    public function getTrace($feedbackID)
    {
        $this->loadModel('action');
        $feedback  = $this->getById($feedbackID);
        $users     = $this->loadModel('user')->getPairs('noletter');
        $actions   = $this->dao->select('*')->from(TABLE_ACTION)->where('objectType')->eq('feedback')->andWhere('objectID')->eq($feedbackID)->orderBy('date')->fetchAll('id');
        $treemap   = "<ul class='tree' id='feedbackTree'>";
        $preAction = '';
        foreach($actions as $action)
        {
            if($preAction == "{$action->actor}-{$action->action}") continue;
            $preAction = "{$action->actor}-{$action->action}";

            if(isset($this->lang->action->label->{$action->action})) $actionName = $this->lang->action->label->{$action->action};
            $user = zget($users, $action->actor);
            $treemap .= "<li class='item-feedback'><span class='title'>{$user}</span><span class='label label-action'>{$actionName}</span><span class='label label-type'>{$this->lang->feedback->common}</span><span class='label label-id'>#{$feedbackID}</span>";
            if($action->extra and is_numeric($action->extra))
            {
                $table = '';
                if($action->action == 'tostory') $table = TABLE_STORY;
                if($action->action == 'tobug')   $table = TABLE_BUG;
                if($action->action == 'totodo')  $table = TABLE_TODO;
                if($action->action == 'totask')  $table = TABLE_TASK;
                if($table)
                {
                    $objectType    = substr($action->action, 2);
                    $object        = $this->dao->select('*')->from($table)->where('id')->eq($action->extra)->fetch();
                    $objectActions = $this->dao->select('*')->from(TABLE_ACTION)->where('objectType')->eq($objectType)->andWhere('objectID')->eq($action->extra)->orderBy('date')->fetchAll('id');
                    $subPreAction  = '';
                    $this->app->loadLang($objectType);

                    $treemap .= "<ul>";
                    foreach($objectActions as $objectAction)
                    {
                        if($objectAction->action == 'opened' or $objectAction->action == 'fromfeedback') continue;
                        if($subPreAction == "{$objectAction->actor}-{$objectAction->action}") continue;
                        $subPreAction = "{$objectAction->actor}-{$objectAction->action}";

                        if(isset($this->lang->action->label->{$objectAction->action})) $actionName = $this->lang->action->label->{$objectAction->action};
                        $user     = zget($users, $objectAction->actor);
                        $treemap .= "<li class='item-{$objectType}'><span class='title'>{$user}</span><span class='label label-action'>{$actionName}</span><span class='label label-type'>{$this->lang->$objectType->common}</span><span class='label label-id'>#{$objectAction->objectID}</span>";

                        $tasks = array();
                        if($objectType == 'story' and $objectAction->action == 'linked2project')
                        {
                            $tasks = $this->dao->select('*')->from(TABLE_TASK)->where('story')->eq($objectAction->objectID)->andWhere('project')->eq($objectAction->extra)->andWhere('deleted')->eq('0')->fetchAll('id');
                        }
                        elseif($objectType == 'bug' and $objectAction->action == 'totask')
                        {
                            $tasks = $this->dao->select('*')->from(TABLE_TASK)->where('fromBug')->eq($objectAction->objectID)->fetchAll('id');
                        }
                        if($tasks)
                        {
                            $this->app->loadLang('task');
                            $taskActions = $this->dao->select('*')->from(TABLE_ACTION)->where('objectType')->eq('task')->andWhere('objectID')->in(array_keys($tasks))->orderBy('date')->fetchGroup('objectID', 'id');
                            foreach($tasks as $taskID => $task)
                            {
                                $treemap .= '<ul>';
                                foreach($taskActions[$taskID] as $taskAction)
                                {
                                    if(isset($this->lang->action->label->{$taskAction->action})) $actionName = $this->lang->action->label->{$taskAction->action};
                                    $user     = zget($users, $taskAction->actor);
                                    $treemap .= "<li class='item-task'><span class='title'>{$user}</span><span class='label label-action'>{$actionName}</span><span class='label label-type'>{$this->lang->task->common}</span><span class='label label-id'>#{$taskID}</span>";
                                }
                                $treemap .= '</ul>';
                            }
                        }

                        $treemap .= '</li>';
                    }
                    $treemap .= '</ul>';
                }
            }
            $treemap .= '</li>';
        }
        $treemap .= '</ul>';
        return $treemap;
    }

    public function manageProduct($productID)
    {
        $this->dao->delete()->from(TABLE_FEEDBACKVIEW)->where('product')->eq($productID)->exec();
        $accounts = fixer::input('post')->remove('allchecker')->get('accounts');
        foreach($accounts as $account)
        {
            $view = new stdclass();
            $view->account = $account;
            $view->product = $productID;
            $this->dao->replace(TABLE_FEEDBACKVIEW)->data($view)->exec();
        }
    }

    public function close($feedbackID)
    {
        $feedback = $this->dao->select('*')->from(TABLE_FEEDBACK)->where('id')->eq($feedbackID)->fetch();
        if($feedback->status != 'closed' and ($feedback->openedBy == $this->app->user->account or empty($this->app->user->feedback)))
        {
            $data = new stdclass();
            $data->status       = 'closed';
            $data->closedBy     = $this->app->user->account;
            $data->closedDate   = helper::now();
            $data->closedReason = $this->post->closedReason;
            $this->dao->update(TABLE_FEEDBACK)->data($data)->where('id')->eq($feedbackID)->exec();

            $changes = common::createChanges($feedback, $data);
            if($changes)
            {
                $actionID = $this->loadModel('action')->create('feedback', $feedbackID, 'closed', $this->post->comment, $this->post->closedReason);
                $this->action->logHistory($actionID, $changes);
            }

        }
    }

    public function printCell($value, $feedback, $users, $allProducts, $depts, $modulePairs, $viewMethod = 'view', $stories = array(), $bugs = array(), $todos = array(), $tasks = array())
    {
        $canBatchEdit     = common::hasPriv('feedback', 'batchEdit');
        $canBatchClose    = common::hasPriv('feedback', 'batchClose');
        $canBatchAssignTo = common::hasPriv('feedback', 'batchAssignTo');
        $canBatchAction   = $canBatchEdit or $canBatchClose or $canBatchAssignTo;

        $feedbackLink = inlink($viewMethod, "feedbackID=$feedback->id");
        $adminMethod  = $this->app->getMethodName() == 'admin';
        $account = $this->app->user->account;
        $user    = $this->loadModel('user')->getByID($account);
        $id      = $value->id;
        $title   = '';
        $style   = '';
        $class   = "c-$id";
        if($id == 'product')    $title = "title='" . zget($allProducts, $feedback->product) . "'";
        if($id == 'title')      $title = "title='{$feedback->title}'";
        if($id == 'actions')    $class.= " text-left";
        if($id == 'status' and $feedback->status == 'clarify')    $class.= " text-red";

        echo "<td class='" . $class . "' $title $style>";
        $this->loadModel('flow')->printFlowCell('feedback', $feedback, $id);
        switch($id)
        {
        case 'id':
            if($canBatchAction)
            {
                echo html::checkbox('feedbackIDList', array($feedback->id => '')) . html::a(helper::createLink('feedback', 'view', "feedbackID=$feedback->id"), sprintf('%03d', $feedback->id));
            }
            else
            {
                printf('%03d', $feedback->id);
            }
            break;
        case 'product':
            echo zget($allProducts, $feedback->product);
            break;
        case 'title':
            if(isset($modulePairs[$feedback->module])) echo "<span class='label label-gray label-badge'>{$modulePairs[$feedback->module]}</span> ";
            echo html::a($feedbackLink, $feedback->title);
            break;
        case 'status':
            echo $this->processStatus('feedback', $feedback);
            break;
        case 'dept':
            echo zget($depts, $user->dept, '/');
            break;
        case 'openedBy':
            echo zget($users, $feedback->openedBy);
            break;
        case 'openedDate':
            echo substr($feedback->openedDate, 5, 11);
            break;
        case 'assignedTo':
            $this->printAssignedHtml($feedback, $users);
            break;
        case 'processedBy':
            echo zget($users, $feedback->processedBy);
            break;
        case 'processedDate':
            echo substr($feedback->processedDate, 5, 11);
            break;
        case 'closedReason':
            if($feedback->closedReason)
            {
                echo zget($this->lang->feedback->closedReasonList, $feedback->closedReason);
                if($feedback->closedReason == 'tostory' and isset($stories[$feedback->result])) echo " #{$feedback->result} <span class='label label-info'>" . zget($this->lang->story->stageList, $stories[$feedback->result]->stage) . "</span>";
                if($feedback->closedReason == 'tobug'   and isset($bugs[$feedback->result]))    echo " #{$feedback->result} <span class='label label-info'>" . $this->processStatus('bug', $bugs[$feedback->result]) . "</span>";
                if($feedback->closedReason == 'totodo'  and isset($todos[$feedback->result]))   echo " #{$feedback->result} <span class='label label-info'>" . zget($this->lang->todo->statusList, $todos[$feedback->result]->status) . "</span>";
                if($feedback->closedReason == 'totask'  and isset($tasks[$feedback->result]))   echo " #{$feedback->result} <span class='label label-info'>" . $this->processStatus('task', $tasks[$feedback->result]) . "</span>";
            }
            break;
        case 'actions':
            $self = $feedback->openedBy == $this->app->user->account;
            if(self::isClickable($feedback, 'edit'))
            {
                echo html::a(inlink('edit', "id=$feedback->id"), "<i class='icon-edit'></i>", '', "class='btn' title='{$this->lang->feedback->edit}'");
            }
            else
            {
                echo html::a('javascript:;', "<i class='icon-edit'></i>", '', "class='btn disabled' style='color:#ccc' title='{$this->lang->feedback->edit}'");
            }

            if($adminMethod)
            {
                if($feedback->status == 'noreview')
                {
                    common::printIcon('feedback', 'review', "feedbackID=$feedback->id", $feedback, 'list', 'glasses', '', 'iframe', true, '', $this->lang->feedback->review);
                }
                else
                {
                    common::printIcon('feedback', 'comment', "feedbackID=$feedback->id&type=replied", $feedback, 'list', 'restart', '', 'iframe', true, '', $this->lang->feedback->reply);
                    if(self::isClickable($feedback, 'comment'))
                    {
                        if($this->config->global->flow == 'full' or $this->config->global->flow == 'onlyStory') common::printIcon('story', 'create', "product=$feedback->product&branch=0&moduleID=0&storyID=0&projectID=0&bugID=0&planID=0&todoID=0&extra=feedbackID=$feedback->id", '', 'list', $this->lang->icons['story'], '', '', '', '', $this->lang->feedback->toStory);
                        if($this->config->global->flow == 'full' or $this->config->global->flow == 'onlyTask') common::printIcon('task', 'create', "projectID=0&storyID=0&moduleID=0&taskID=0&extra=feedbackID=$feedback->id", '', 'list', $this->lang->icons['task'], '', '', '', '', $this->lang->feedback->toTask);
                        if($this->config->global->flow == 'full' or $this->config->global->flow == 'onlyTest')  common::printIcon('bug',   'create', "product=$feedback->product&branch=0&extras=feedbackID=$feedback->id", '', 'list', $this->lang->icons['bug'], '', '', '', '', $this->lang->feedback->toBug);
                        if($this->config->global->flow == 'full')  common::printIcon('todo', 'create', "date=today&account=&from=feedback&feedbackID=$feedback->id", '', 'list', $this->lang->icons['todo'], '', '', '', '', $this->lang->feedback->toTodo);
                    }
                    else
                    {
                        if($this->config->global->flow == 'full' or $this->config->global->flow == 'onlyStory') echo html::a('javascript:;', "<i class='icon-{$this->lang->icons['story']}'></i>", '', "class='btn disabled' title='{$this->lang->feedback->toStory}'");
                        if($this->config->global->flow == 'full' or $this->config->global->flow == 'onlyTask') echo html::a('javascript:;', "<i class='icon-{$this->lang->icons['task']}'></i>", '', "class='btn disabled' title='{$this->lang->feedback->toTask}'");
                        if($this->config->global->flow == 'full' or $this->config->global->flow == 'onlyTest') echo html::a('javascript:;', "<i class='icon-{$this->lang->icons['bug']}'></i>", '', "class='btn disabled' title='{$this->lang->feedback->toBug}'");
                        if($this->config->global->flow == 'full')  echo html::a('javascript:;', "<i class='icon-{$this->lang->icons['todo']}'></i>", '', "class='btn disabled' title='{$this->lang->feedback->toTodo}'");
                    }
                }
            }

            common::printIcon('feedback', 'close', "feedbackID=$feedback->id", $feedback, 'list', 'off', '', 'iframe', '', '', $this->lang->feedback->close);
            common::printIcon('feedback', 'delete', "feedbackID=$feedback->id", $feedback, 'list', 'trash', 'hiddenwin', '', '', '', $this->lang->feedback->delete);
            break;
            echo '</td>';
        }
    }

    public function printAssignedHtml($feedback, $users)
    {
        $btnTextClass   = '';
        $hasPriv        = common::hasPriv('feedback', 'assignTo');
        $assignedToText = !empty($feedback->assignedTo) ? zget($users, $feedback->assignedTo) : $this->lang->feedback->noAssigned;
        $btnTextClass   = $hasPriv ? 'text-primary' : '';
        if($feedback->assignedTo == $this->app->user->account) $btnTextClass = 'text-red';

        $btnClass     = $feedback->assignedTo == 'closed' ? ' disabled' : '';
        $btnClass     = "iframe btn btn-icon-left btn-sm {$btnClass}";

        $assignToLink = helper::createLink('feedback', 'assignTo', "feedbackID=$feedback->id", '', true);
        $assignToHtml = html::a($assignToLink, "<i class='icon icon-hand-right'></i> <span title='$feedback->assignedTo' class='{$btnTextClass}'>{$assignedToText}</span>", '', "class='$btnClass'");

        echo !$hasPriv ? "<span style='padding-left: 21px' class='{$btnTextClass}'>{$assignedToText}</span>" : $assignToHtml;
    }
    
    public function sendmail($feedbackID, $actionID, $to = '')
    {
        $feedback = $this->getByID($feedbackID);
        if(!$feedback->notify) return false;
        $users = $this->loadModel('user')->getPairs('noletter');

        $this->loadModel('mail');
		$productName = $this->loadModel('product')->getById($feedback->product)->name;

        /* Get action info. */
        $action = $this->loadModel('action')->getById($actionID);

        $history            = $this->action->getHistory($actionID);
        $action->history    = isset($history[$actionID]) ? $history[$actionID] : array();
        $action->appendLink = '';
        if(strpos($action->extra, ':') !== false)
        {
            list($extra, $id) = explode(':', $action->extra);
            $action->extra    = $extra;

            $table = '';
            if($action->objectType == 'bug')   $table = TABLE_BUG;
            if($action->objectType == 'story') $table = TABLE_STORY;
            if($id and $table)
            {
                $name  = $this->dao->select('title')->from($table)->where('id')->eq($id)->fetch('title');
                if($name) $action->appendLink = html::a(zget($this->config->mail, 'domain', common::getSysURL()) . helper::createLink($action->objectType, 'view', "id=$id", 'html'), "#$id " . $name);
            }
        }

        /* Get mail content. */
        $modulePath = $this->app->getModulePath($appName = '', 'feedback');
        $oldcwd     = getcwd();
        $viewFile   = $modulePath . 'view/sendmail.html.php';
        chdir($modulePath . 'view');
        if(file_exists($modulePath . 'ext/view/sendmail.html.php'))
        {
            $viewFile = $modulePath . 'ext/view/sendmail.html.php';
            chdir($modulePath . 'ext/view');
        }
        ob_start();
        include $viewFile;
        foreach(glob($modulePath . 'ext/view/sendmail.*.html.hook.php') as $hookFile) include $hookFile;
        $mailContent = ob_get_contents();
        ob_end_clean();
        chdir($oldcwd);

        /* Set toList and ccList. */

        $sendUsers = $this->getToAndCcList($feedback);
        if(!$sendUsers) return;
        list($toList, $ccList) = $sendUsers;
        
        if($to != 'asked')  $ccList .= ',' . $feedback->openedBy;

        /* Send it. */
        $this->mail->send($toList, $this->lang->feedback->common . ' #'. $feedback->id . ' ' . $feedback->title . ' - ' . $productName, $mailContent, $ccList);
        
        if($this->mail->isError()) trigger_error(join("\n", $this->mail->getError()));
    }

    public function getToAndCcList($feedback)
    {
        /* Set toList and ccList. */
        $toList = $feedback->assignedTo;

        $ccList = '';
        if(!$feedback->notify) $ccList .= $feedback->openedBy . ',';
        if($feedback->mailto)  $ccList .= $feedback->mailto . ',';
        if($feedback->product)
        {
            $product = $this->loadModel('product')->getByID($feedback->product);
            if($product->feedback) $ccList .= $product->feedback . ',';
        }
        $ccList = trim(str_replace(",{$toList},", ',', ",{$ccList},"), ',');

        if(empty($toList))
        {
            if(empty($ccList)) return false;
            if(strpos($ccList, ',') === false)
            {
                $toList = $ccList;
                $ccList = '';
            }
            else
            {
                $commaPos = strpos($ccList, ',');
                $toList = substr($ccList, 0, $commaPos);
                $ccList = substr($ccList, $commaPos + 1);
            }
        }

        return array($toList, $ccList);
    }

    public function forceNotReview()
    {    
        if(empty($this->config->feedback->needReview))
        {    
            if(!isset($this->config->feedback->forceReview)) return true;
            if(strpos(",{$this->config->feedback->forceReview},", ",{$this->app->user->account},") === false) return true;
        }    
        if($this->config->feedback->needReview && strpos(",{$this->config->feedback->forceNotReview},", ",{$this->app->user->account},") !== false) return true;

        return false;
    }

    public static function isClickable($feedback, $action)
    {
        global $app, $config;
        $action = strtolower($action);
        if($action == 'edit' and strpos('wait|noreview|clarify', $feedback->status) !== false and ($app->user->account == $feedback->openedBy or $app->user->admin)) return true;
        if($action == 'delete') return true;
        if($action == 'assignto' and empty($app->user->feedback)) return true;
        if($action == 'review' and empty($app->user->feedback) and $feedback->status == 'noreview') return true;
        if($action == 'comment' and empty($app->user->feedback) and strpos('closed|clarify|noreview', $feedback->status) === false) return true;
        if($action == 'close' and $feedback->status != 'closed' and ($app->user->account == $feedback->openedBy or empty($app->user->feedback))) return true;

        return false;
    }

    /**
     * Get status for different method.
     *
     * @param  string $methodName
     * @param  bool   $needReview
     * @param  string $status
     * @access public
     * @return string
     */
    public function getStatus($methodName, $needReview, $status = '')
    {
        if($methodName == 'create') $status = $needReview ? 'noreview' : 'wait';
        if($methodName == 'update')
        {
            if($status == 'clarify') $status = $needReview ? 'noreview' : 'wait';
        }
        if($methodName == 'review')
        {
            if($this->post->result == 'pass')    $status = 'wait';
            if($this->post->result == 'clarify') $status = 'clarify';
        }
        return $status;
    }
}
